local errors = require('errors')
local labels = require('ouroboros-cartridge-sidecar.metric_labels')

local _ = package.search('libouroboros') or
    package.searchpath('libouroboros', package.cpath) or
    error('Failed to load libouroboros, check LUA_CPATH')
local role_name = 'ouroboros-replicator'

local RustError = errors.new_class("RustError")

local EXT_FN_NAMES = { 'apply_dml_replace', 'apply_dml_delete', 'apply_ddl', 'test_lib' }

local function init(opts)
    if opts.is_master then
        labels.add()
    end

    -- luacheck: push ignore
    local module = box.lib.load('libouroboros')
    -- luacheck: pop
    if not module then
        error('Can not load module "libouroboros"')
    end
    local _, err = RustError:pcall(module:load("init"))
    if err then
        return nil, err
    end

    for _, func_name in ipairs(EXT_FN_NAMES) do
        local func = module:load(func_name)
        if not func then
            error('Function ' .. func_name .. ' cannot be loaded from libouroboros')
        end
        rawset(_G, 'libouroboros_' .. func_name, function(...) return func(...) end)
    end
end

local function stop()
    labels.remove()
end

local function validate_config()
    return true
end

local was_master = false
local function apply_config(_, options)
    if options.is_master and not was_master then
        labels.add()
    end
    if not options.is_master and was_master then
        labels.remove()
    end
    was_master = options.is_master
end

local function get_issues()
    local issues = {}

    if not labels.metrics_role_found() then
        table.insert(issues, {
            level = 'warn',
            topic = 'ouroboros-cartridge-sidecar.replicator',
            message = 'ouroboros-cartridge-sidecar.replicator: Роль `metrics` не найдена'
        })
    end
    return issues
end

return {
    role_name = role_name,
    init = init,
    stop = stop,
    validate_config = validate_config,
    apply_config = apply_config,
    get_issues = get_issues,
    dependencies = {
        'cartridge.roles.vshard-storage',
        'cartridge.roles.metrics'
    },
}
