local fun = require('fun')
local msgpack = require('msgpack')
local net_box = require('net.box')
local t = require('luatest')
local g = t.group('ouroboros_topology')

local helper = require('test.helper.integration')

local requests = require('test.requests')

g.test_version = function()
    local ver = require('ouroboros-cartridge-sidecar.version')
    local resp = requests.http_request('GET', 'version')
    t.assert_equals(resp.status, 200)
    t.assert(resp.json, 'Response is not json')
    local data = resp.json
    t.assert_equals(data.version, ver)
end

g.test_topology = function()
    local resp = requests.http_request('GET', 'topology')
    t.assert_equals(resp.status, 200)
    t.assert(resp.json, 'Response is not json')
    local data = resp.json
    t.assert_equals(#data, 3)
    -- vshard group 'default'
    t.assert_equals(data[1].name, 'default')
    t.assert(data[1].replicasets)
    t.assert_equals(#data[1].replicasets, 1)
    t.assert_equals(data[1].replicasets[1].vshard_group, 'default')
    t.assert_equals(data[1].replicasets[1].alias, 'topology')
    t.assert(data[1].replicasets[1].spaces)
    t.assert_equals(#data[1].replicasets[1].servers, 1)
    t.assert(data[1].replicasets[1].http_address)
    -- vshard group 'test'
    t.assert_equals(data[2].name, 'test')
    t.assert_equals(data[2].bucket_count, 1000)
    t.assert_equals(data[2].bucket_idx_name, 'bucket_id')
    t.assert(data[2].replicasets)
    t.assert_equals(#data[2].replicasets, 1)
    t.assert_equals(data[2].replicasets[1].vshard_group, 'test')
    t.assert_equals(data[2].replicasets[1].alias, 'replicaset-1')
    t.assert(data[2].replicasets[1].spaces)
    t.assert_equals(#data[2].replicasets[1].spaces, 2)
    t.assert_equals(data[2].replicasets[1].spaces[1], { bucket_id_index = 0, space_id = 513, space_name = "space_one" })
    t.assert_equals(data[2].replicasets[1].spaces[2], { bucket_id_index = 2, space_id = 514, space_name = "space_two" })
    t.assert_equals(#data[2].replicasets[1].servers, 2)
    t.assert_not(data[2].replicasets[1].http_address)
    -- vshard group 'new'
    t.assert_equals(data[3].name, 'new')
    t.assert_equals(data[3].bucket_count, 1000)
    t.assert_equals(data[3].bucket_idx_name, 'bucket_id')
    t.assert(data[3].replicasets)
    t.assert_equals(#data[3].replicasets, 1)
    t.assert_equals(data[3].replicasets[1].vshard_group, 'new')
    t.assert_equals(data[3].replicasets[1].alias, 'replicaset-2')
    t.assert(data[3].replicasets[1].spaces)
    t.assert_equals(#data[3].replicasets[1].spaces, 1)
    t.assert_equals(data[3].replicasets[1].spaces[1], {
        bucket_id_index = 1,
        space_id = 514,
        space_name = "space_four"
    })
    t.assert_equals(#data[3].replicasets[1].servers, 3)
    t.assert_not(data[3].replicasets[1].http_address)
end

g.test_health = function()
    local resp = requests.http_request('GET', 'health')
    t.assert_equals(resp.status, 200)
end

g.test_ouroboros_user = function()
    local topology = require('ouroboros-cartridge-sidecar.roles.ouroboros-topology')
    local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
    local topologies = helper.get_servers_by_role_name(helper.cluster, topology.role_name)
    local replicators = helper.get_servers_by_role_name(helper.cluster, replicator.role_name)

    local not_exist = function(acc, server)
        return acc or server.net_box:eval('return box.space._vuser.index.name:get("ouroboros") ~= nil')
    end
    local exist = function(acc, server)
        return acc and server.net_box:eval('return box.space._vuser.index.name:get("ouroboros") ~= nil')
    end

    local before_on_topologies = false
    local res = fun.foldl(not_exist, before_on_topologies, topologies)
    t.assert_not(res)

    local before_on_replicators = false
    res = fun.foldl(not_exist, before_on_replicators, replicators)
    t.assert_not(res)

    local resp = requests.http_request('POST', 'user', {})
    t.assert_equals(resp.status, 400)

    local password = 'some_pass'
    resp = requests.http_request('POST', 'user', { body = { new_pass = password } })
    t.assert_equals(resp.status, 200)

    local after_on_topologies = false
    res = fun.foldl(not_exist, after_on_topologies, topologies)
    t.assert_not(res)

    local after_on_replicators = true
    res = fun.foldl(exist, after_on_replicators, replicators)
    t.assert(res)

    local server = helper.get_master_by_role_name(helper.cluster, replicator.role_name)
    local conn = net_box.connect(server.advertise_uri, { user = 'ouroboros', password = password })
    t.assert(conn, 'Error on connecting to replicator as user')
    local cres, cerr = conn:eval("return box.space._space:count() > 0") -- user must have access to universe
    t.assert_not(cerr, ('%s'):format(cerr))
    t.assert(cres)
    cres, cerr = conn:call('libouroboros_test_lib') -- user must have access to stored funcs
    t.assert_not(cerr, ('%s'):format(cerr))
    t.assert(cres)
    conn:close()
end

g.test_space_info = function()
    local resp = requests.http_request('GET', 'space')
    t.assert_equals(resp.status, 404)

    resp = requests.http_request('GET', 'space/some')
    t.assert_equals(resp.status, 404)

    resp = requests.http_request('GET', 'space/some/some')
    t.assert_equals(resp.status, 404)
    t.assert(resp.json, 'Response is not json')
    t.assert(resp.json.error, 'Response has no error message')

    -- unknown space
    resp = requests.http_request('GET', 'space/new/space')
    t.assert_equals(resp.status, 200)
    t.assert_equals(resp.json, {})

    -- non-sharded space
    resp = requests.http_request('GET', 'space/new/space_three')
    t.assert_equals(resp.status, 200)
    t.assert_equals(resp.json, {})

    resp = requests.http_request('GET', 'space/new/space_four')
    t.assert_equals(resp.status, 200)
    t.assert(resp.json.space, 'Space info is not returned')
    t.assert_equals(type(resp.json.space), 'string')
    local info = string.fromhex(resp.json.space)
    local val, pos = msgpack.decode(info)
    t.assert_equals(type(val), 'table')
    t.assert_gt(pos, info:len())
    t.assert(resp.json.indexes, 'Space indexes is not returned')
    t.assert_equals(#resp.json.indexes, 2)

    -- space_two from vshard group 'test'
    resp = requests.http_request('GET', 'space/test/514')
    t.assert_equals(resp.status, 200)
    t.assert(resp.json.space, 'Space info is not returned')
    t.assert_equals(type(resp.json.space), 'string')
    info = string.fromhex(resp.json.space)
    val, pos = msgpack.decode(info)
    t.assert_equals(type(val), 'table')
    t.assert_gt(pos, info:len())
    t.assert(resp.json.indexes, 'Space indexes is not returned')
    t.assert_equals(#resp.json.indexes, 1)
end
